from time import perf_counter

from panther.request import Request
from panther.logger import monitoring_logger
from panther.middlewares.base import BaseMiddleware


class Middleware(BaseMiddleware):
    """
    Create Log Message Like Below:
    [method] path | ip:port | response_time ms | status_code
    """

    async def before(self, request: Request) -> Request:
        ip, port = request.client
        self.log = f'{request.method} | {request.path} | {ip}:{port}'
        self.start_time = perf_counter()
        return request

    async def after(self, status_code: int):
        """
        We handled Monitoring Middle manually,
        cause of that we only have "status_code" here
        """
        response_time = (perf_counter() - self.start_time) * 1_000
        monitoring_logger.info(f'{self.log} | {response_time: .3} ms | {status_code}')



from typing import Self

from panther.db.connection import db
from panther.db.utils import merge_dicts
from panther.exceptions import DBException


class BasePantherDBQuery:

    @classmethod
    def _merge(cls, *args) -> dict:
        # TODO: Convert "id" to "_id"
        return merge_dicts(*args)

    # # # # # Find # # # # #
    @classmethod
    def find_one(cls, _data: dict = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).find_one(**cls._merge(_data, kwargs)):
            return cls(**document)

    @classmethod
    def find(cls, _data: dict = None, /, **kwargs) -> list[Self]:
        documents = db.session.collection(cls.__name__).find(**cls._merge(_data, kwargs))
        return [cls(**document) for document in documents]

    # # # # # Insert # # # # #
    @classmethod
    def insert_one(cls, _data: dict = None, **kwargs) -> Self:
        document = db.session.collection(cls.__name__).insert_one(**cls._merge(_data, kwargs))
        return cls(**document)

    @classmethod
    def insert_many(cls, _data: dict = None, /, **kwargs):
        raise DBException('insert_many() is not supported while using PantherDB.')

    # # # # # Delete # # # # #
    def delete(self) -> None:
        db.session.collection(self.__class__.__name__).delete_one(_id=self.id)

    @classmethod
    def delete_one(cls, _data: dict = None, /, **kwargs) -> bool:
        return db.session.collection(cls.__name__).delete_one(**cls._merge(_data, kwargs))

    @classmethod
    def delete_many(cls, _data: dict = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).delete_many(**cls._merge(_data, kwargs))

    # # # # # Update # # # # #
    def update(self, **kwargs) -> None:
        for field, value in kwargs.items():
            setattr(self, field, value)
        db.session.collection(self.__class__.__name__).update_one({'_id': self.id}, **kwargs)

    @classmethod
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> bool:
        return db.session.collection(cls.__name__).update_one(_filter, **cls._merge(_data, kwargs))

    @classmethod
    def update_many(cls, _filter, _data: dict = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).update_many(_filter, **cls._merge(_data, kwargs))

    # # # # # Other # # # # #
    @classmethod
    def count(cls, _data: dict = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).count(**cls._merge(_data, kwargs))



import asyncio
from pathlib import Path
from runpy import run_path
from pydantic.main import ModelMetaclass

from panther import status
from panther.request import Request
from panther.response import Response
from panther.exceptions import APIException
from panther.configs import JWTConfig, config
from panther.middlewares.base import BaseMiddleware
from panther.middlewares.monitoring import Middleware as MonitoringMiddleware
from panther.routings import find_endpoint, check_urls, collect_urls, finalize_urls
from panther._utils import http_response, import_class, read_body, collect_path_variables

""" We can't import logger on the top cause it needs config['base_dir'] ans its fill in __init__ """


class Panther:

    def __init__(self, name):
        import os
        os.system('clear')
        self.base_dir = Path(name).resolve().parent
        config['base_dir'] = self.base_dir
        self.panther_dir = Path(__file__).parent
        self.load_configs()
        del os

    async def __call__(self, scope, receive, send) -> None:
        """
        We Used Python3.11 For asyncio.TaskGroup()
        1.
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.run(scope, receive, send))
        2.
            await self.run(scope, receive, send)
        3.
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(self.run, scope, receive, send)
                await anyio.to_thread.run_sync(self.run, scope, receive, send)
        4.
            with ProcessPoolExecutor() as e:
                e.submit(self.run, scope, receive, send)
        """
        async with asyncio.TaskGroup() as tg:
            tg.create_task(self.run(scope, receive, send))

    async def run(self, scope, receive, send):
        from panther.logger import logger
        # Read Body & Create Request
        body = await read_body(receive)
        request = Request(scope=scope, body=body)

        # Monitoring Middleware
        monitoring_middleware = None
        if config['monitoring']:
            monitoring_middleware = MonitoringMiddleware()
            await monitoring_middleware.before(request=request)

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=request.path)
        path_variables = collect_path_variables(request_path=request.path, found_path=found_path)

        if endpoint is None:
            return await http_response(
                send, status_code=status.HTTP_404_NOT_FOUND, monitoring=monitoring_middleware, exception=True,
            )

        try:  # They Both Have The Save Exception (APIException)
            # Call 'Before' Middlewares
            for middleware in config['middlewares']:
                request = await middleware.before(request=request)

            # User Didn't Use API Decorator
            if not hasattr(endpoint, '__wrapped__'):
                logger.critical(f'You may have forgotten to use @API on the {endpoint.__name__}()')
                return await http_response(
                    send,
                    status_code=status.HTTP_510_NOT_EXTENDED,
                    monitoring=monitoring_middleware,
                    exception=True,
                )

            # Call Endpoint
            response = await endpoint(request=request, **path_variables)
        except APIException as e:
            response = self.handle_exceptions(e)
        except Exception as e:
            # Every unhandled exception in Panther or code will catch here
            logger.critical(e)
            return await http_response(
                send,
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                monitoring=monitoring_middleware,
                exception=True,
            )

        # Call 'After' Middleware
        for middleware in config['reversed_middlewares']:
            try:
                response = await middleware.after(response=response)
            except APIException as e:
                response = self.handle_exceptions(e)

        await http_response(
            send, status_code=response.status_code, monitoring=monitoring_middleware, body=response.body,
        )

    @classmethod
    def handle_exceptions(cls, e, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )

    def load_configs(self) -> None:
        from panther.logger import logger
        logger.debug(f'Base directory: {self.base_dir}')

        # Check & Read The Configs File
        self._check_configs()

        # Put Variables In "config"
        config['monitoring'] = self.settings.get('MONITORING', config['monitoring'])
        config['log_queries'] = self.settings.get('LOG_QUERIES', config['log_queries'])
        config['default_cache_exp'] = self.settings.get('DEFAULT_CACHE_EXP', config['default_cache_exp'])
        config['throttling'] = self.settings.get('THROTTLING', config['throttling'])
        config['secret_key'] = self._get_secret_key()

        config['middlewares'] = self._get_middlewares()
        config['reversed_middlewares'] = config['middlewares'][::-1]
        config['user_model'] = self._get_user_model()

        config['authentication'] = self._get_authentication_class()
        config['jwt_config'] = self._get_jwt_config()

        # Check & Collect URLs
        #   check_urls should be the last call in load_configs, because it will read all files and load them.
        urls = check_urls(self.settings.get('URLs')) or {}
        collected_urls = dict()
        collect_urls('', urls, collected_urls)
        config['urls'] = finalize_urls(collected_urls)
        logger.debug('Configs loaded.')
        if config['monitoring']:
            logger.info('Run "panther monitor" in another session for Monitoring.')

    def _check_configs(self):
        from panther.logger import logger
        """Read the config file and put it as dict in self.settings"""
        try:
            configs_path = self.base_dir / 'core/configs.py'
            self.settings = run_path(str(configs_path))
        except FileNotFoundError:
            logger.critical('core/configs.py Not Found.')

    def _get_secret_key(self) -> bytes | None:
        if secret_key := self.settings.get('SECRET_KEY'):
            return secret_key.encode()
        return None

    def _get_authentication_class(self) -> ModelMetaclass | None:
        return self.settings.get('AUTHENTICATION') and import_class(self.settings['AUTHENTICATION'])

    def _get_user_model(self) -> ModelMetaclass:
        return import_class(self.settings.get('USER_MODEL', 'panther.db.models.User'))

    def _get_jwt_config(self) -> JWTConfig:
        """Only Collect JWT Config If Authentication Is JWTAuthentication"""
        if getattr(config['authentication'], '__name__', None) == 'JWTAuthentication':
            user_config = self.settings.get('JWTConfig')
            return JWTConfig(**user_config) if user_config else JWTConfig(key=config['secret_key'].decode())

    def _get_middlewares(self) -> list:
        """Collect The Middlewares & Set db_engine If One Of Middlewares Was For DB"""
        from panther.logger import logger
        middlewares = list()

        for path, data in self.settings.get('MIDDLEWARES', []):
            if path.find('panther.middlewares.db.Middleware') != -1:
                config['db_engine'] = data['url'].split(':')[0]

            # noinspection PyPep8Naming
            Middleware = import_class(path)
            if not issubclass(Middleware, BaseMiddleware):
                logger.critical(f'{Middleware} is not a sub class of BaseMiddleware.')
                continue
            # noinspection PyArgumentList
            middlewares.append(Middleware(**data))
        return middlewares



import orjson as json
from types import NoneType
from datetime import timedelta
from collections import namedtuple


from panther.logger import logger
from panther.configs import config
from panther.request import Request
from panther.db.connection import redis
from panther.response import Response, ResponseDataTypes


caches = dict()
Cached = namedtuple('Cached', ['data', 'status_code'])


def cache_key(request: Request, /):
    client = request.user and request.user.id or request.client.ip
    return f'{client}-{request.path}-{request.data}'


def get_cached_response_data(*, request: Request) -> Cached | None:
    """
    If redis.is_connected:
        Get Cached Data From Redis
    else:
        Get Cached Data From Memory
    """
    key = cache_key(request)
    if redis.is_connected:  # NOQA: Unresolved References
        data = (redis.get(key) or b'{}').decode()
        if cached := json.loads(data):
            return Cached(*cached)
        else:
            return None

    else:
        global caches
        if cached := caches.get(key):
            return Cached(*cached)
        else:
            return None


def set_cache_response(*, request: Request, response: Response, cache_exp_time: timedelta | int) -> None:
    """
    If redis.is_connected:
        Cache The Data In Redis
    else:
        Cache The Data In Memory
    """
    key = cache_key(request)
    cache_data: tuple[ResponseDataTypes, int] = (response._data, response.status_code)

    if redis.is_connected:  # NOQA: Unresolved References
        cache_exp_time = cache_exp_time or config['default_cache_exp']
        cache_data: bytes = json.dumps(cache_data)

        if not isinstance(cache_exp_time, (timedelta, int, NoneType)):
            raise TypeError('cache_exp_time should be "datetime.timedelta" or "int" or "None"')

        if cache_exp_time is None:
            logger.warning(
                'your response are going to cache in redis forever '
                '** set DEFAULT_CACHE_EXP in configs or pass the cache_exp_time in @API.get() for prevent this **'
            )
            redis.set(key, cache_data)
        else:
            redis.set(key, cache_data, ex=cache_exp_time)

    else:
        global caches
        caches[key] = cache_data

        if cache_exp_time:
            logger.error('cache_exp_time not supported while redis is not connected.')



import re
from copy import deepcopy
from runpy import run_path
from typing import Callable
from collections import Counter
from typing import MutableMapping
from collections.abc import Mapping
from functools import reduce, partial


from panther.configs import config


def check_urls(urls: str | None) -> dict | None:
    from panther.logger import logger

    if urls is None:
        return logger.critical("configs.py Does Not Have 'URLs'")

    try:
        full_urls_path = config['base_dir'] / urls
        urls_dict = run_path(str(full_urls_path))['urls']
    except FileNotFoundError:
        return logger.critical("Couldn't Open 'URLs' Address.")
    except KeyError:
        return logger.critical("'URLs' Address Does Not Have 'urls'")
    if not isinstance(urls_dict, dict):
        return logger.critical("'urls' Of URLs Is Not dict.")
    return urls_dict


def collect_urls(pre_url: str, urls: dict, final: dict):
    from panther.logger import logger

    for url, endpoint in urls.items():
        if endpoint is ...:
            logger.error(f"URL Can't Point To Ellipsis. ('{pre_url}{url}' -> ...)")
        elif endpoint is None:
            logger.error(f"URL Can't Point To None. ('{pre_url}{url}' -> None)")
        elif url and not re.match(r'[a-zA-Z<>0-9_/-]', url):
            logger.error(f"URL Is Not Valid. --> '{pre_url}{url}'")
        else:
            if not url.endswith('/'):
                url = f'{url}/'
            if isinstance(endpoint, dict):
                if url != '/':
                    if pre_url:
                        pre_url = f'{pre_url}/{url}'
                    else:
                        pre_url = url
                collect_urls(pre_url, endpoint, final)
            else:
                final[f'{pre_url}{url}'] = endpoint

    return urls


def find_endpoint(path: str) -> tuple[Callable | None, str]:
    if (location := path.find('?')) != -1:
        path = path[:location]
    path = path.removesuffix('/').removeprefix('/')  # 'user/list'
    paths = path.split('/')  # ['user', 'list']
    paths_len = len(paths)
    sub = config['urls']
    # sub = {
    #     'user': {
    #         '<id>': <function users at 0x7f579d060220>,
    #         '': <function single_user at 0x7f579d060e00>
    #     }
    # }
    found_path = ''
    for i, split_path in enumerate(paths):
        last_path = bool((i + 1) == paths_len)
        found = sub.get(split_path)
        if last_path and callable(found):
            found_path += f'{split_path}/'
            return found, found_path
        if isinstance(found, dict):
            found_path += f'{split_path}/'
            if last_path and callable(endpoint := found.get('')):
                return endpoint, found_path

            sub = found
            continue

        # found = None
        # sub = {'<id>': <function return_list at 0x7f0757baff60>} (Example)
        _continue = False
        for key, value in sub.items():
            if key.startswith('<'):
                if last_path:
                    if callable(value):
                        found_path += f'{key}/'
                        return value, found_path
                    else:
                        return None, ''

                sub = value
                found_path += f'{key}/'
                _continue = True
                break

        if _continue:
            continue

    return None, ''


def is_recursive_merge(a, b):
    both_mapping = isinstance(a, Mapping) and isinstance(b, Mapping)
    both_counter = isinstance(a, Counter) and isinstance(b, Counter)
    return both_mapping and not both_counter


def deepmerge(dst, src):
    for key in src:
        if key in dst:
            if is_recursive_merge(dst[key], src[key]):
                deepmerge(dst[key], src[key])
            else:
                dst[key] = deepcopy(src[key])
        else:
            dst[key] = deepcopy(src[key])
    return dst


def merge(destination: MutableMapping, *sources) -> MutableMapping:
    """Credit to Travis Clarke --> https://github.com/clarketm/mergedeep"""
    return reduce(partial(deepmerge), sources, destination)


def finalize_urls(urls: dict):
    urls_list = list()
    for url, endpoint in urls.items():
        path = dict()
        for single_path in url.split('/')[:-1][::-1]:
            path = {single_path: path or endpoint}
        urls_list.append(path)
    return merge(*urls_list)



from pathlib import Path
from typing import TypedDict
from datetime import timedelta
from dataclasses import dataclass
from pydantic.main import ModelMetaclass

from panther.throttling import Throttling


@dataclass(frozen=True)
class JWTConfig:
    key: str
    algorithm: str = 'HS256'
    life_time: timedelta | int = timedelta(days=1)


class Config(TypedDict):
    base_dir: Path
    monitoring: bool
    log_queries: bool
    urls: dict
    middlewares: list
    reversed_middlewares: list
    db_engine: str
    default_cache_exp: timedelta | None
    secret_key: bytes | None
    authentication: ModelMetaclass | None
    jwt_config: JWTConfig | None
    user_model: ModelMetaclass | None
    throttling: Throttling | None


config: Config = {
    'base_dir': Path(),
    'monitoring': False,
    'log_queries': False,
    'secret_key': None,
    'urls': {},
    'middlewares': [],
    'reversed_middlewares': [],
    'db_engine': '',
    'default_cache_exp': None,
    'jwt_config': None,
    'authentication': None,
    'user_model': None,
    'throttling': None,
}



from panther.logger import logger

logo = r"""│    ____                 __    __                         │
│   /\  _`\              /\ \__/\ \                        │
│   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     │
│    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   │
│     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    │
│      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    │
│       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    │
"""

help_message = f"""╭{58 * '─'}╮
{logo}│{58 * ' '}│
│                                                          │
│   usage:                                                 │
│       - panther create <project_name> <directory>        │
│           Create your project in current directory       │
│                                                          │
│       - panther run [--reload | --help]                  │
│           Run your project with uvicorn                  │
│                                                          │
│       - panther shell                                    │
│           Run interactive python shell                   │
│                                                          │
│       - panther monitor                                  │
│           Show the monitor :)                            │
│                                                          │
│       - panther [--help | -h ]                           │
│           Show this message and exit                     │
╰{58 * '─'}╯
"""

run_help_message = """
    --host TEXT                     Bind socket to this host.  [default: 127.0.0.1]
    --port INTEGER                  Bind socket to this port.  [default: 8000]
    --uds TEXT                      Bind to a UNIX domain socket.
    --fd INTEGER                    Bind to socket from this file descriptor.
    --reload                        Enable auto-reload.
    --reload-dir PATH               Set reload directories explicitly, instead
                                    of using the current working directory.
    --reload-include TEXT           Set glob patterns to include while watching
                                    for files. Includes '*.py' by default; these
                                    defaults can be overridden with `--reload-
                                    exclude`. This option has no effect unless
                                    watchfiles is installed.
    --reload-exclude TEXT           Set glob patterns to exclude while watching
                                    for files. Includes '.*, .py[cod], .sw.*,
                                    ~*' by default; these defaults can be
                                    overridden with `--reload-include`. This
                                    option has no effect unless watchfiles is
                                    installed.
    --reload-delay FLOAT            Delay between previous and next check if
                                    application needs to be. Defaults to 0.25s.
                                    [default: 0.25]
    --workers INTEGER               Number of worker processes. Defaults to the
                                    $WEB_CONCURRENCY environment variable if
                                    available, or 1. Not valid with --reload.
    --loop [auto|asyncio|uvloop]    Event loop implementation.  [default: auto]
    --http [auto|h11|httptools]     HTTP protocol implementation.  [default: auto]
    --ws [auto|none|websockets|wsproto]
                                    WebSocket protocol implementation. [default: auto]
    --ws-max-size INTEGER           WebSocket max size message in bytes [default: 16777216]
    --ws-ping-interval FLOAT        WebSocket ping interval  [default: 20.0]
    --ws-ping-timeout FLOAT         WebSocket ping timeout  [default: 20.0]
    --ws-per-message-deflate BOOLEAN
                                    WebSocket per-message-deflate compression default: True]
    --lifespan [auto|on|off]        Lifespan implementation.  [default: auto]
    --interface [auto|asgi3|asgi2|wsgi]
                                    Select ASGI3, ASGI2, or WSGI as the
                                    application interface.  [default: auto]
    --env-file PATH                 Environment configuration file.
    --log-config PATH               Logging configuration file. Supported 
                                        formats: .ini, .json, .yaml.
    --log-level [critical|error|warning|info|debug|trace]
                                    Log level. [default: info]
    --access-log / --no-access-log  Enable/Disable access log.
    --use-colors / --no-use-colors  Enable/Disable colorized logging.
    --proxy-headers / --no-proxy-headers
                                    Enable/Disable X-Forwarded-Proto, X-Forwarded-For, 
                                    X-Forwarded-Port to populate remote address info.
    --server-header / --no-server-header
                                    Enable/Disable default Server header.
    --date-header / --no-date-header
                                    Enable/Disable default Date header.
    --forwarded-allow-ips TEXT      Comma separated list of IPs to trust with
                                    proxy headers. Defaults to the
                                    $FORWARDED_ALLOW_IPS environment variable if
                                    available, or '127.0.0.1'.
    --root-path TEXT                Set the ASGI 'root_path' for applications
                                    sub mounted below a given URL path.
    --limit-concurrency INTEGER     Maximum number of concurrent connections or
                                    tasks to allow, before issuing HTTP 503 responses.
    --backlog INTEGER               Maximum number of connections to hold in backlog
    --limit-max-requests INTEGER    Maximum number of requests to service before
                                    terminating the process.
    --timeout-keep-alive INTEGER    Close Keep-Alive connections if no new data
                                    is received within this timeout.  [default: 5]
    --ssl-keyfile TEXT              SSL key file
    --ssl-certfile TEXT             SSL certificate file
    --ssl-keyfile-password TEXT     SSL keyfile password
    --ssl-version INTEGER           SSL version to use (see stdlib ssl module's)
                                    [default: 17]
    --ssl-cert-reqs INTEGER         Whether client certificate is required 
                                    (see stdlib ssl module's)  [default: 0]
    --ssl-ca-certs TEXT             CA certificates file
    --ssl-ciphers TEXT              Ciphers to use (see stdlib ssl module's)
                                    [default: TLSv1]
    --header TEXT                   Specify custom default HTTP response headers
                                    as a Name:Value pair
    --version                       Display the uvicorn version and exit.
    --app-dir TEXT                  Look for APP in the specified directory, by
                                    adding this to the PYTHONPATH. Defaults to
                                    the current working directory.  [default: .]
    --h11-max-incomplete-event-size INTEGER
                                    For h11, the maximum number of bytes to
                                    buffer of an incomplete event.
    --factory                       Treat APP as an application factory, i.e. a
                                    () -> <ASGI app> callable.
    --help                          Show this message and exit.
    """


def cli_error(message: str | TypeError) -> None:
    logger.error(f'Error: {message}\n\nUse panther -h for more help')


def import_error_message(module_name: str) -> str:
    return f'No module named "{module_name}"\n\t    Hint: Try to install with "pip install {module_name}"'


def import_error(module_name: str) -> None:
    logger.critical(import_error_message(module_name))


def clean_args(args: list[str]) -> dict:
    """
    Input: ['--reload', '--host', '127.0.0.1', ...]
    Output: {'--reload: None, 'host': '127.0.0.1', ...}
    """
    _args = dict()
    for i, arg in enumerate(args):
        if arg.startswith('--'):
            if (i + 1) < len(args):
                _args[arg[2:]] = args[i + 1]
            else:
                _args[arg[2:]] = True
    return _args



import orjson as json
from types import NoneType
from pydantic.main import BaseModel as PydanticBaseModel

ResponseDataTypes = int | dict | list | tuple | set | str | bool | NoneType
IterableDataTypes = list | tuple | set


class Response:
    def __init__(self, data: ResponseDataTypes = None, status_code: int = 200):
        """
        :param data: should be int | dict | list | tuple | set | str | bool or NoneType
        :param status_code: should be int
        """
        # TODO: Handle bytes data
        data = self.clean_data_type(data)
        self.check_status_code(status_code)

        self._status_code = status_code
        self._data = data
        # TODO: Add Header To Response

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def body(self) -> bytes:
        return json.dumps(self._data)

    def set_data(self, data) -> None:
        self._data = data

    @classmethod
    def check_status_code(cls, status_code: any):
        if not isinstance(status_code, int):
            error = f'Response "status_code" Should Be "int". ("{status_code}" -> {type(status_code)})'
            raise TypeError(error)

    @classmethod
    def clean_data_type(cls, data: any):
        """
        Make sure the response data is only ResponseDataTypes or Iterable of ResponseDataTypes
        """
        if issubclass(type(data), PydanticBaseModel):
            return data.dict()

        elif isinstance(data, IterableDataTypes):
            return [cls.clean_data_type(d) for d in data]

        else:
            return data



from panther.request import Request


class BasePermission:
    """
    Just for demonstration
    """
    @classmethod
    def authorization(cls, request: Request) -> bool:
        return True


class AdminPermission:

    @classmethod
    def authorization(cls, request: Request) -> bool:
        return request.user and hasattr(request.user, 'is_admin') and request.user.is_admin



import os
import sys

from rich import print as rprint

from panther.cli.run_command import run
from panther.cli.create_command import create
from panther.cli.monitor_command import monitor
from panther.cli.utils import clean_args, help_message, cli_error


def shell() -> None:
    os.system('bpython')


def start() -> None:
    command = sys.argv and sys.argv[1] or None
    args = clean_args(sys.argv[2:])

    match command:
        case '-h' | '--help':
            rprint(help_message)
        case 'create':
            create(sys.argv[2:])
        case 'run':
            run(args)
        case 'shell':
            shell()
        case 'monitor':
            monitor()
        case _:
            cli_error('Invalid Arguments.')



import os
import logging
from pydantic import BaseModel

from logging.config import dictConfig
from panther.configs import config


LOGS_DIR = config['base_dir'] / 'logs'
if not os.path.exists(LOGS_DIR):
    os.makedirs(LOGS_DIR)


class LogConfig(BaseModel):
    """Logging configuration to be set for the server."""

    LOGGER_NAME: str = 'panther-logger'
    DEFAULT_LOG_FORMAT: str = '%(levelprefix)s | %(asctime)s | %(message)s'
    FILE_LOG_FORMAT: str = '%(asctime)s | %(message)s'
    LOG_LEVEL: str = 'DEBUG'
    MAX_FILE_SIZE: int = 1024 * 1024 * 100  # 100 MB

    version = 1
    disable_existing_loggers = False

    formatters = {
        'default': {
            '()': 'uvicorn.logging.DefaultFormatter',
            'fmt': DEFAULT_LOG_FORMAT,
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'file_formatter': {
            '()': 'uvicorn.logging.DefaultFormatter',
            'fmt': FILE_LOG_FORMAT,
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
    }
    handlers = {
        'monitoring_file': {
            'formatter': 'file_formatter',
            'filename': LOGS_DIR / 'monitoring.log',
            'class': 'logging.handlers.RotatingFileHandler',
            'maxBytes': MAX_FILE_SIZE,  # 100 MB,
            'backupCount': 3,
        },
        'query_file': {
            'formatter': 'file_formatter',
            'filename': LOGS_DIR / 'query.log',
            'class': 'logging.handlers.RotatingFileHandler',
            'maxBytes': MAX_FILE_SIZE,  # 100 MB,
            'backupCount': 3,
        },
        'file': {
            'formatter': 'file_formatter',
            'filename': LOGS_DIR / f'main.log',
            'class': 'logging.handlers.RotatingFileHandler',
            'maxBytes': MAX_FILE_SIZE,  # 100 MB,
            'backupCount': 3,
        },
        'default': {
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
        },
    }
    loggers = {
        'panther': {
            'handlers': ['default', 'file'],
            'level': LOG_LEVEL,
        },
        'monitoring': {
            'handlers': ['monitoring_file'],
            'level': LOG_LEVEL,
        },
        'query': {
            'handlers': ['default', 'query_file'],
            'level': LOG_LEVEL,
        },
    }


dictConfig(LogConfig().dict())
logger = logging.getLogger('panther')
query_logger = logging.getLogger('query')
monitoring_logger = logging.getLogger('monitoring')



import os
import sys
import time

from panther.cli.template import Template
from panther.cli.utils import cli_error


def create(args: list):
    # Get Project Name
    if len(args) == 0:
        return cli_error('Not Enough Parameters.')
    project_name = args[0]

    # Get Base Directory
    base_directory = project_name
    if len(args) > 1:
        base_directory = args[1]

    # Check All The Directories Existence
    existence = check_all_directories(base_directory)
    if existence:
        return cli_error(f'"{existence}" Directory Already Exists.')

    load_animation()

    # Create Base Directory
    if base_directory != '.':
        os.makedirs(base_directory)

    for file_name, data in Template.items():
        if isinstance(data, dict):
            # Create Sub Directory
            sub_directory = f'{base_directory}/{file_name}'
            os.makedirs(sub_directory)

            # Create Files of Sub Directory
            for sub_file_name, sub_data in data.items():
                file_path = f'{sub_directory}/{sub_file_name}'
                sub_data = sub_data.replace('{PROJECT_NAME}', project_name.lower())
                with open(file_path, 'x') as file:
                    file.write(sub_data)
        else:
            # Create File
            data = data.replace('{PROJECT_NAME}', project_name.lower())
            file_path = f'{base_directory}/{file_name}'
            with open(file_path, 'x') as file:
                file.write(data)
    else:
        print('Project Created Successfully.')


def check_all_directories(base_directory: str) -> str | None:
    """return folder_name means that the directory exist."""
    if base_directory != '.' and os.path.isdir(base_directory):
        return base_directory

    for file_name, data in Template.items():
        sub_directory = f'{base_directory}/{file_name}'
        if os.path.exists(sub_directory):
            return sub_directory

        if isinstance(data, dict):
            for sub_file_name, _ in data.items():
                file_path = f'{sub_directory}/{sub_file_name}'
                if os.path.exists(file_path):
                    return file_path


def load_animation():
    animation = [
        '■□□□□□□□□□□',
        '■■□□□□□□□□□',
        '■■■□□□□□□□□',
        '■■■■□□□□□□□',
        '■■■■■□□□□□□',
        '■■■■■■□□□□□',
        '■■■■■■■□□□□',
        '■■■■■■■■□□□',
        '■■■■■■■■■□□',
        '■■■■■■■■■■□',
        '■■■■■■■■■■■',
    ]

    for i in range(len(animation)):
        time.sleep(0.2)
        sys.stdout.write('\r' + 'Creating Your Project: ' + animation[i % len(animation)])
        sys.stdout.flush()

    print('\n')



import re
import importlib

import orjson as json

from panther.logger import logger
from panther.status import status_text


async def read_body(receive) -> bytes:
    """Read and return the entire body from an incoming ASGI message."""
    body = b''
    more_body = True
    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)
    return body


async def _http_response_start(send, /, status_code: int):
    await send({
        'type': 'http.response.start',
        'status': status_code,
        'headers': [
            [b'content-type', b'application/json'],
        ],
    })


async def _http_response_body(send, /, body: any = None):
    if body is None:
        await send({'type': 'http.response.body'})
    else:
        await send({'type': 'http.response.body', 'body': body})


async def http_response(
        send,
        /,
        *,
        status_code: int,
        monitoring,  # type: MonitoringMiddleware | None
        body: bytes = None,
        exception: bool = False
):
    if exception:
        body = json.dumps({'detail': status_text[status_code]})
    elif status_code == 204 or body == b'null':
        body = None

    if monitoring is not None:
        await monitoring.after(status_code=status_code)

    await _http_response_start(send, status_code=status_code)
    await _http_response_body(send, body=body)


def import_class(_klass: str, /):
    """
    Example:
        Input: panther.db.models.User
        Output: User (The Class)
    """
    seperator = _klass.rfind('.')
    module = importlib.import_module(_klass[:seperator])
    return getattr(module, _klass[seperator + 1:])


def read_multipart_form_data(content_type: str, body: str) -> dict:
    """
    content_type = 'application/json'
    content_type = 'multipart/form-data; boundary=--------------------------984465134948354357674418'
    """
    boundary = content_type[30:]

    per_pattern = r'(.*\r\nContent-Disposition: form-data; name=")(.*)'  # (Junk)(FieldName)
    value_pattern = r'(\r\n\r\n)(.*)'  # (Junk)(Value)

    # (Junk)(FieldName) (Junk)(Value)(Junk)
    field_pattern = per_pattern + value_pattern + r'(\r\n.*)'

    # (Junk)(FieldName) (Junk)(FileName)(Junk)(ContentType) (Junk)(Value)
    file_pattern = per_pattern + r'("; filename=")(.*)("\r\nContent-Type: )(.*)' + value_pattern

    fields = dict()
    for field in body.split(boundary):
        if match := re.match(pattern=field_pattern, string=field):
            _, field_name, _, value, _ = match.groups()
            fields[field_name] = value

        if match := re.match(pattern=file_pattern, string=field):
            # TODO: It works but it is not profitable, So comment it for later
            #   We should handle it while we are reading the body in _utils.read_body()

            # _, field_name, _, file_name, _, content_type, _, value = match.groups()
            # data = {
            #     'filename': file_name,
            #     'Content-Type': content_type,
            #     'value': value
            # }
            # fields[field_name] = data
            logger.error("We Don't Handle Files In Multipart Request Yet.")
    return fields


def collect_path_variables(request_path: str, found_path: str) -> dict:
    found_path = found_path.removesuffix('/').removeprefix('/')
    request_path = request_path.removesuffix('/').removeprefix('/')
    path_variables = dict()
    for f_path, r_path in zip(found_path.split('/'), request_path.split('/')):
        if f_path.startswith('<'):
            path_variables[f_path[1:-1]] = r_path
    return path_variables



import bson
import operator
from functools import reduce
from time import perf_counter

from panther.configs import config
from panther.logger import query_logger


def log_query(func):
    def log(*args, **kwargs):
        if config['log_queries'] is False:
            return func(*args, **kwargs)
        start = perf_counter()
        response = func(*args, **kwargs)
        end = perf_counter()
        class_name = args[0].__name__ if hasattr(args[0], '__name__') else args[0].__class__.__name__
        query_logger.info(f'\033[1mQuery -->\033[0m  {class_name}.{func.__name__}() --> {(end - start) * 1_000:.2} ms')
        return response
    return log


def clean_object_id(_id: bson.ObjectId | str) -> bson.ObjectId:
    if isinstance(_id, bson.ObjectId):
        return _id
    try:
        return bson.ObjectId(_id)
    except Exception:
        raise bson.errors.InvalidId


def clean_object_id_in_dicts(*args):
    # TODO: Refactor this func
    for d in args:
        if d is None:
            continue
        if '_id' in d:
            d['_id'] = clean_object_id(d['_id'])
        if 'id' in d:
            d['_id'] = clean_object_id(d.pop('id'))


def merge_dicts(*args) -> dict:
    return reduce(operator.ior, filter(None, args), {})



import functools
from datetime import timedelta, datetime

from pydantic import ValidationError
from orjson.orjson import JSONDecodeError

from panther import status
from panther.logger import logger
from panther.configs import config
from panther.request import Request
from panther.throttling import Throttling, throttling_storage
from panther.response import Response, IterableDataTypes
from panther.caching import get_cached_response_data, set_cache_response, cache_key
from panther.exceptions import APIException, InvalidPathVariableException, AuthorizationException, JsonDecodeException, \
    ThrottlingException
from panther.utils import round_datetime


class API:
    def __init__(
            self,
            input_model=None,
            output_model=None,
            auth: bool = False,
            permissions: list | None = None,
            throttling: Throttling = None,
            cache: bool = False,
            cache_exp_time: timedelta | int | None = None,
    ):
        self.input_model = input_model
        self.output_model = output_model
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.cache_exp_time = cache_exp_time
        self.request: Request | None = None

    def __call__(self, func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            self.request: Request = kwargs.pop('request')  # NOQA: Non-self attribute could not be type hinted

            # Handle Authentication
            self.handle_authentications()

            # Handle Throttling
            self.handle_throttling()

            # Handle Authentication
            self.handle_permissions()

            # Validate Input
            self.validate_input()

            # Validate Path Variables
            self.validate_path_variables(func, kwargs)

            # Get Cached Response
            if self.cache and self.request.method == 'GET':
                if cached := get_cached_response_data(request=self.request):
                    return Response(data=cached.data, status_code=cached.status_code)

            # Put Request In kwargs
            if req_arg := [k for k, v in func.__annotations__.items() if v == Request]:
                kwargs[req_arg[0]] = self.request

            # Call Endpoint
            response = await func(**kwargs)

            # Clean Output
            if not isinstance(response, Response):
                response = Response(data=response)
            data = self.serialize_response_data(data=response._data)  # NOQA: Access to a protected member
            response.set_data(data)

            # Set New Response To Cache
            if self.cache and self.request.method == 'GET':
                set_cache_response(request=self.request, response=response, cache_exp_time=self.cache_exp_time)

            # Warning CacheExpTime
            if self.cache_exp_time and self.cache is False:
                logger.warning('"cache_exp_time" won\'t work while "cache" is False')

            return response
        return wrapper

    def handle_authentications(self) -> None:
        auth_class = config['authentication']
        if self.auth:
            if not auth_class:
                raise TypeError('"AUTHENTICATION" has not been set in core/configs')
            user = auth_class.authentication(self.request)
            self.request.set_user(user=user)

    def handle_throttling(self) -> None:
        if throttling := self.throttling or config['throttling']:
            key = cache_key(self.request)
            time = round_datetime(datetime.now(), throttling.duration)
            throttling_key = f'{time}-{key}'
            if throttling_storage[throttling_key] > throttling.rate:
                raise ThrottlingException

            throttling_storage[throttling_key] += 1

    def handle_permissions(self) -> None:
        for perm in self.permissions:
            if type(perm.authorization).__name__ != 'method':
                logger.error(f'{perm.__name__}.authorization should be "classmethod"')
                continue
            if perm.authorization(request=self.request) is False:
                raise AuthorizationException

    def validate_input(self):
        if self.input_model:
            try:
                validated_data = self.input_model(**self.request.pure_data)
                self.request.set_validated_data(validated_data)
            except ValidationError as validation_error:
                error = {e['loc'][0]: e['msg'] for e in validation_error.errors()}
                raise APIException(status_code=status.HTTP_400_BAD_REQUEST, detail=error)
            except JSONDecodeError:
                raise JsonDecodeException

    def serialize_response_data(self, data):
        """
        We serializer the response here instead of response.py file
        because we don't have access to the "output_model" there
        """

        # None or Unchanged
        if data is None or self.output_model is None:
            return data

        return self.serialize_with_output_model(data)

    def serialize_with_output_model(self, data: any):
        # Dict
        if isinstance(data, dict):
            return self.output_model(**data).dict()

        # Iterable
        elif isinstance(data, IterableDataTypes):
            return [self.serialize_with_output_model(d) for d in data]

        # Str | Bool
        raise TypeError('Type of Response data is not match with output_model. '
                        '\n*hint: You may want to pass None to output_model')

    @staticmethod
    def validate_path_variables(func, variables):
        for user_var, value in variables.items():
            for func_var, _type in func.__annotations__.items():
                if user_var == func_var:
                    if _type is bool:
                        variables[user_var] = value.lower() not in ['false', '0']
                    elif _type is int:
                        try:
                            variables[user_var] = int(value)
                        except ValueError:
                            raise InvalidPathVariableException(value=value, arg_type=_type)
                    break



from datetime import datetime
from panther.logger import logger
from panther.configs import config
from panther.db.models import User
from panther.request import Request
from panther.cli.utils import import_error
from panther.exceptions import AuthenticationException
try:
    from jose import JWTError, jwt
except ImportError:
    import_error('python-jose')
    exit()


class JWTAuthentication:
    model = User
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # Header encoding (see RFC5987)

    @staticmethod
    def get_authorization_header(request: Request):
        auth = request.headers.authorization
        if isinstance(auth, str):
            auth = auth.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        return auth

    @classmethod
    def authentication(cls, request: Request):
        auth = cls.get_authorization_header(request).split()
        if not auth or auth[0].lower() != cls.keyword.lower().encode():
            logger.error(f'JWT Authentication Error: "token keyword is not valid"')
            raise AuthenticationException
        if len(auth) != 2:
            logger.error(f'JWT Authentication Error: "len of token should be 2"')
            raise AuthenticationException
        try:
            token = auth[1].decode()
        except UnicodeError:
            logger.error(f'JWT Authentication Error: "Unicode Error"')
            raise AuthenticationException

        payload = cls.decode_jwt(token)
        return cls.get_user(payload)

    @classmethod
    def get_user(cls, payload: dict):
        if (user_id := payload.get('user_id')) is None:
            logger.error(f'JWT Authentication Error: "Payload does not have user_id"')
            raise AuthenticationException

        user_model = config['user_model'] or cls.model
        if user := user_model.find_one(id=user_id):
            return user

        logger.error(f'JWT Authentication Error: "User not found"')
        raise AuthenticationException

    @staticmethod
    def encode_jwt(user_id: int) -> str:
        """Encode JWT from user_id."""
        expire = datetime.utcnow() + config['jwt_config'].life_time
        access_payload = {
            'token_type': 'access',
            'user_id': user_id,
            'exp': expire,
        }
        return jwt.encode(access_payload, config['jwt_config'].key, algorithm=config['jwt_config'].algorithm)

    @staticmethod
    def decode_jwt(token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(token, config['jwt_config'].key, algorithms=[config['jwt_config'].algorithm])
        except JWTError as e:
            logger.error(f'JWT Authentication Error: "{e}"')
            raise AuthenticationException

    @classmethod
    def login(cls, user_id: int) -> str:
        """alias of encode_jwt()"""
        return cls.encode_jwt(user_id=user_id)

    @staticmethod
    def logout(user_id: int):
        # TODO:
        #   1. Save the token in cache
        #   2. Check logout cached token in authentication()
        pass



from redis import Redis
from pantherdb import PantherDB
from panther.configs import config
from panther.cli.utils import import_error_message


class Singleton(object):
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instances[cls]


class DBSession(Singleton):
    _db_name: str

    def __init__(self, db_url: str | None = None):
        if db_url:
            if (seperator := db_url.find(':')) == -1:
                # ex: db_url = 'some_db' (or something without ':')
                seperator = None
            self._db_name = db_url[:seperator]
            match self._db_name:
                case 'mongodb':
                    self._create_mongodb_session(db_url)
                case 'pantherdb':
                    self._create_pantherdb_session(db_url[12:])
                case _:
                    raise ValueError(f'We are not support "{self._db_name}" database yet')

    @property
    def session(self):
        return self._session

    @property
    def name(self) -> str:
        return self._db_name

    def _create_mongodb_session(self, db_url: str) -> None:
        try:
            from pymongo import MongoClient
            from pymongo.database import Database
        except ImportError:
            raise ImportError(import_error_message('pymongo'))

        self._client: MongoClient = MongoClient(db_url)
        self._session: Database = self._client.get_database()

    def _create_pantherdb_session(self, db_url: str):
        self._session: PantherDB = PantherDB(db_url, return_dict=True, secret_key=config['secret_key'])

    def close(self):
        if self._db_name == 'mongodb':
            self._client.close()
        else:
            self._session.close()


class RedisConnection(Singleton, Redis):
    is_connected: bool = False

    def __init__(self, host: str | None = None, port: int | None = None, **kwargs):
        if host and port:
            super().__init__(host=host, port=port, **kwargs)
            self.is_connected = True


db: DBSession = DBSession()
redis: Redis = RedisConnection()



import os
import uvicorn
from rich import print as rprint
from panther.cli.utils import cli_error, run_help_message


def _handle_boolean_commands(args: dict[str, str | None]) -> dict:
    """Boolean Commands:
    - reload
    - access-log
    - no-access-log
    - use-colors
    - no-use-colors
    - server-header
    - no-server-header
    """
    _command = dict()

    if 'reload' in args:
        _command['reload'] = bool(args.pop('reload', None))

    _command['access_log'] = False  # Default
    if 'access-log' in args:
        _command['access_log'] = bool(args.pop('access-log', None))

    if 'no-access-log' in args:
        _command['access_log'] = not bool(args.pop('no-access-log', None))

    if 'use-colors' in args:
        _command['use_colors'] = bool(args.pop('use-colors', None))

    if 'no-use-colors' in args:
        _command['use_colors'] = not bool(args.pop('no-use-colors', None))

    if 'server-header' in args:
        _command['server_header'] = bool(args.pop('server-header', None))

    if 'no-server-header' in args:
        _command['server_header'] = not bool(args.pop('no-server-header', None))

    return _command


def run(args: dict[str, str | None]) -> None:
    if 'h' in args or 'help' in args:
        rprint(run_help_message)
        return
    command = {'app_dir': os.getcwd()}
    command.update(_handle_boolean_commands(args))
    command.update(args)

    try:
        uvicorn.run('main:app', **command)
    except TypeError as e:
        cli_error(e)



from typing import Self

from panther.db.connection import db  # NOQA: F401
from panther.exceptions import DBException
from panther.db.utils import clean_object_id_in_dicts, merge_dicts


class BaseMongoDBQuery:

    @classmethod
    def _merge(cls, *args) -> dict:
        clean_object_id_in_dicts(*args)
        # TODO: Convert "id" to "_id"
        return merge_dicts(*args)

    # # # # # Find # # # # #
    @classmethod
    def find_one(cls, _data: dict = None, /, **kwargs) -> Self | None:
        _query = cls._merge(_data, kwargs)
        if document := eval(f'db.session.{cls.__name__}.find_one(_query)'):
            return cls(**document)

    @classmethod
    def find(cls, _data: dict = None, /, **kwargs) -> list[Self]:
        _query = cls._merge(_data, kwargs)
        documents = eval(f'db.session.{cls.__name__}.find(_query)')
        return [cls(**document) for document in documents]

    # # # # # Insert # # # # #
    @classmethod
    def insert_one(cls, _data: dict = None, **kwargs) -> Self:
        document = cls._merge(_data, kwargs)
        document['id'] = eval(f'db.session.{cls.__name__}.insert_one(document)').inserted_id
        return cls(**document)

    @classmethod
    def insert_many(cls, _data: dict = None, **kwargs) -> Self:
        raise DBException('insert_many() is not supported while using MongoDB.')

    # # # # # Delete # # # # #
    def delete(self) -> None:
        _filter = {'_id': self._id}
        eval(f'db.session.{self.__class__.__name__}.delete_one(_filter)')

    @classmethod
    def delete_one(cls, _data: dict = None, /, **kwargs) -> bool:
        _query = cls._merge(_data, kwargs)
        result = eval(f'db.session.{cls.__name__}.delete_one(_query)')
        return bool(result.deleted_count)

    @classmethod
    def delete_many(cls, _data: dict = None, /, **kwargs) -> int:
        _query = cls._merge(_data, kwargs)
        result = eval(f'db.session.{cls.__name__}.delete_many(_query)')
        return result.deleted_count

    # # # # # Update # # # # #
    def update(self, **kwargs) -> None:
        for field, value in kwargs.items():
            setattr(self, field, value)
        _filter = {'_id': self._id}
        _update = {'$set': kwargs}
        eval(f'db.session.{self.__class__.__name__}.update_one(_filter, _update)')

    @classmethod
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> bool:
        clean_object_id_in_dicts(_filter)
        _update = {'$set': kwargs | {}}
        if isinstance(_data, dict):
            _data['$set'] = _data.get('$set', {}) | (kwargs or {})

        result = eval(f'db.session.{cls.__name__}.update_one(_filter, _data | _update)')
        return bool(result.updated_count)

    @classmethod
    def update_many(cls, _filter, _data: dict = None, /, **kwargs) -> int:
        _update = {'$set': cls._merge(_data, kwargs)}
        result = eval(f'db.session.{cls.__name__}.update_many(_filter, _update)')
        return result.updated_count

    # # # # # Other # # # # #
    @classmethod
    def count(cls, _data: dict = None, /, **kwargs) -> int:
        _query = cls._merge(_data, kwargs)
        return eval(f'db.session.{cls.__name__}.count_documents(_query)')



import os
from datetime import datetime, timedelta
from pathlib import Path
from panther.logger import logger


def load_env(env_file: str | Path, /) -> dict[str, str]:
    variables = dict()

    if env_file is None or not os.path.isfile(env_file):
        logger.critical(f'"{env_file}" is not valid file for load_env()')
        return variables

    with open(env_file) as file:
        for line in file.readlines():
            line = line.strip()
            if '=' in line and not line.startswith('#'):
                key, value = line.split('=', 1)
                key = key.strip()
                value = value.strip().strip('"\'')
                variables[key] = value
    return variables


def generate_secret_key() -> str:
    from cryptography.fernet import Fernet
    return Fernet.generate_key().decode()


def round_datetime(dt: datetime, delta: timedelta):
    return datetime.min + round((dt - datetime.min) / delta) * delta



from panther.db.connection import RedisConnection
from panther.logger import logger
from panther.middlewares.base import BaseMiddleware
from panther.request import Request
from panther.response import Response


class Middleware(BaseMiddleware):

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.validate_host()
        self.validate_port()

    def validate_host(self):
        if host := self.kwargs.get('host'):
            if False:  # TODO: should be valid ip or domain
                logger.critical('Redis "host" is not valid.')
        else:
            self.kwargs['host'] = '127.0.0.1'

    def validate_port(self):
        if port := self.kwargs.get('port'):
            if isinstance(port, str):
                try:
                    self.kwargs['port'] = int(port)
                except ValueError:
                    logger.critical('Redis "port" should be number')

            elif not isinstance(port, int):
                logger.critical('Redis "port" is not valid.')
        else:
            self.kwargs['port'] = 6379

    async def before(self, request: Request) -> Request:
        self.redis = RedisConnection(**self.kwargs)
        return request

    async def after(self, response: Response) -> Response:
        self.redis.close()
        return response



import os
from collections import deque

from rich import box
from rich.align import Align
from rich.console import Console, Group
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from watchfiles import watch

from panther.cli.utils import cli_error


def monitor() -> None:
    def _generate_table(rows):
        layout = Layout()
        console = Console()

        rows = list(rows)
        n_rows = os.get_terminal_size()[1]

        while n_rows >= 0:
            table = Table(box=box.MINIMAL_DOUBLE_HEAD)
            table.add_column('Datetime', justify='center', style='magenta', no_wrap=True)
            table.add_column('Method', justify='center', style='cyan')
            table.add_column('Path', justify='center', style='cyan')
            table.add_column('Client', justify='center', style='cyan')
            table.add_column('Response Time', justify='center', style='blue')
            table.add_column('Status Code', justify='center', style='blue')

            for row in rows[-n_rows:]:
                table.add_row(*row)
            layout.update(table)
            render_map = layout.render(console, console.options)

            if len(render_map[layout].render[-1]) > 2:
                n_rows -= 1  # The table is overflowing
            else:
                break

        return Panel(
            Align.center(Group(table)),
            box=box.ROUNDED,
            padding=(1, 2),
            title='Monitoring',
            border_style='bright_blue',
        )

    try:
        with open('logs/monitoring.log') as f:
            f.readlines()
            width, height = os.get_terminal_size()
            messages = deque(maxlen=height - 8)  # Save space for header and footer

            with Live(_generate_table(messages), auto_refresh=False, vertical_overflow='visible', screen=True) as live:
                # TODO: Is it only watch logs/monitoring.log or the whole directory ?
                for _ in watch('logs/monitoring.log'):
                    data = f.readline().split('|')
                    messages.append(data)
                    live.update(_generate_table(messages))
                    live.refresh()

    except FileNotFoundError:
        cli_error("Monitor Log File Does Not Exists.\n\nHint: Make sure 'Monitor' is True in 'core/configs' "
                  "or you are in a correct directory.")
    except KeyboardInterrupt:
        pass



from panther.request import Request
from panther.response import Response


class BaseMiddleware:

    async def before(self, request: Request):
        return request

    async def after(self, response: Response):
        return response



from typing import Self

from pydantic import ValidationError

from panther.configs import config
from panther.db.utils import log_query
from panther.db.queries.mongodb_queries import BaseMongoDBQuery
from panther.db.queries.pantherdb_queries import BasePantherDBQuery
from panther.exceptions import DBException

if config['db_engine'] == 'pantherdb':
    BaseQuery = BasePantherDBQuery
else:
    BaseQuery = BaseMongoDBQuery

__all__ = (
    'Query',
)


class Query(BaseQuery):

    @classmethod
    def validate_data(cls, data: dict, is_updating: bool = False) -> Self | None:
        """
        *. Validate the input of user with its class
        *. If is_updating is True & exception happens but the message was empty
        """
        try:
            cls(**data)
        except ValidationError as validation_error:
            error = ', '.join(
                '{field}="{error}"'.format(field=e['loc'][0], error=e['msg'])
                for e in validation_error.errors() if not is_updating or e['type'] != 'value_error.missing')
            if error:
                message = f'{cls.__name__}({error})'
                raise DBException(message)

    # # # # # Find # # # # #
    @classmethod
    @log_query
    def find_one(cls, _data: dict = None, /, **kwargs) -> Self | None:
        """
        example:
            >>> from example.app.models import User
            >>> User.find(id=1)
        """
        return super().find_one(_data, **kwargs)

    @classmethod
    @log_query
    def find(cls, _data: dict = None, /, **kwargs) -> list[Self]:
        """
        example:
            >>> from example.app.models import User
            >>> User.find(name='Ali')
        """
        return super().find(_data, **kwargs)

    # # # # # Insert # # # # #
    @classmethod
    @log_query
    def insert_one(cls, _data: dict = None, /, **kwargs) -> Self:
        """
        example:
            >>> from example.app.models import User
            >>> User.insert_one(name='Ali', age=24, ...)
        """
        cls.validate_data(kwargs)
        return super().insert_one(_data, **kwargs)

    @classmethod
    @log_query
    def insert_many(cls, _data: dict = None, **kwargs):
        return super().insert_many(_data, **kwargs)

    # # # # # Delete # # # # #
    @log_query
    def delete(self) -> None:
        """
        example:
            >>> from example.app.models import User
            >>> user = User.find_one(name='Ali')
            >>> user.delete()
        """
        return super().delete()

    @classmethod
    @log_query
    def delete_one(cls, **kwargs) -> bool:
        """
        example:
            >>> from example.app.models import User
            >>> User.delete_one(id=1)
        """
        return super().delete_one(**kwargs)

    @classmethod
    @log_query
    def delete_many(cls, **kwargs) -> int:
        """
        example:
            >>> from example.app.models import User
            >>> User.delete_many(last_name='Rn')
        """
        return super().delete_many(**kwargs)

    # # # # # Update # # # # #
    @log_query
    def update(self, **kwargs) -> None:
        """
        example:
            >>> from example.app.models import User
            >>> user = User.find_one(name='Ali')
            >>> user.update(name='Saba')
        """
        self.validate_data(kwargs, is_updating=True)
        return super().update(**kwargs)

    @classmethod
    @log_query
    def update_one(cls, _filter, _data: dict = None, /, **kwargs) -> bool:
        """
        example:
            >>> from example.app.models import User
            >>> User.update_one({'id': 1}, name='Ali')
            >>> User.update_one({'id': 2}, {'name': 'Ali', 'age': 25})
        """
        return super().update_one(_filter, _data, **kwargs)

    @classmethod
    @log_query
    def update_many(cls, _filter, **kwargs) -> int:
        """
        example:
            >>> from example.app.models import User
            >>> User.update_many({'name': 'Mohsen'}, name='Ali')
        """
        return super().update_many(_filter, **kwargs)

    # # # # # Other # # # # #
    @classmethod
    @log_query
    def last(cls, _data: dict = None, /, **kwargs) -> Self | None:
        """
        example:
            >>> from example.app.models import User
            >>> user = User.last(name='Ali')
        """
        return super().last(_data, **kwargs)

    @classmethod
    @log_query
    def count(cls, _data: dict = None, /, **kwargs) -> int:
        """
        example:
            >>> from example.app.models import User
            >>> User.count(name='Ali')
        """
        return super().count(_data, **kwargs)

    @classmethod
    @log_query
    def find_or_insert(cls, **kwargs) -> tuple[bool, any]:
        """
        example:
            >>> from example.app.models import User
            >>> user = User.find_or_insert(name='Ali')
        """
        if obj := cls.find_one(**kwargs):
            return False, obj
        else:
            return True, cls.insert_one(**kwargs)



from collections import defaultdict
from dataclasses import dataclass
from datetime import timedelta

throttling_storage = defaultdict(int)


@dataclass(repr=False, eq=False, match_args=False)
class Throttling:
    rate: int
    duration: timedelta



from panther import status


class DBException(Exception):
    pass


class APIException(Exception):
    detail: str | dict | list = 'Internal Server Error'
    status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(self, detail=None, status_code=None):
        self.detail = detail or self.detail
        self.status_code = status_code or self.status_code


class AuthenticationException(APIException):
    detail = 'Authentication Error'
    status_code = status.HTTP_401_UNAUTHORIZED


class AuthorizationException(APIException):
    detail = 'Permission Denied'
    status_code = status.HTTP_403_FORBIDDEN


class JsonDecodeException(APIException):
    detail = 'JSON Decode Error'
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY


class ThrottlingException(APIException):
    detail = 'Too Many Request'
    status_code = status.HTTP_429_TOO_MANY_REQUESTS


class InvalidPathVariableException(APIException):
    def __init__(self, value: str, arg_type: type):
        detail = f"Path variable '{value}' should be '{arg_type.__name__}'"
        super().__init__(detail=detail, status_code=status.HTTP_400_BAD_REQUEST)




import orjson as json
from typing import Literal
from dataclasses import dataclass
from collections import namedtuple

from panther._utils import read_multipart_form_data


@dataclass(frozen=True)
class Headers:
    accept_encoding: str
    content_length: int
    authorization: str
    content_type: str
    user_agent: str
    connection: str
    accept: str
    host: str


Address = namedtuple('Client', ['ip', 'port'])


class Request:
    def __init__(self, scope: dict, body: bytes):
        """{'type': 'http', 'asgi': {'version': '3.0', 'spec_version': '2.3'},
        'http_version': '1.1', 'server': ('127.0.0.1', 8000), 'client': ('127.0.0.1', 35064),
        'scheme': 'http', 'root_path': '', 'headers': [
            (b'content-type', b'application/json'),
            (b'user-agent', b'PostmanRuntime/7.29.2'),
            (b'accept', b'*/*'),
            (b'postman-token', b'3e78fbf3-df2f-41bd-bedc-cf6027fa4fe6'),
            (b'host', b'127.0.0.1:8000'),
            (b'accept-encoding', b'gzip, deflate, br'),
            (b'connection', b'keep-alive'),
            (b'content-length', b'55')
        ],
        'method': 'GET', 'path': '/list/', 'raw_path': b'/list/', 'query_string': b''}.
        """
        self.scope = scope
        self._body = body
        self._data = None
        self._validated_data = None
        self._user = None
        self._headers: Headers | None = None
        self._params: dict | None = None

    @property
    def headers(self):
        _headers = {header[0].decode('utf-8'): header[1].decode('utf-8') for header in self.scope['headers']}
        if self._headers is None:
            self._headers = Headers(
                accept_encoding=_headers.pop('accept-encoding', None),
                content_length=_headers.pop('content_length', None),
                authorization=_headers.pop('authorization', b''),
                content_type=_headers.pop('content-type', None),  # Handle content-type & boundary together
                user_agent=_headers.pop('user-agent', None),
                connection=_headers.pop('connection', None),
                accept=_headers.pop('accept', None),
                host=_headers.pop('host', None),
                # TODO: Others ...
            )
        return self._headers

    @property
    def query_params(self):
        query_string = self.scope['query_string'].decode('utf-8').split('&')
        if self._params is None:
            self._params = dict()
            for param in query_string:
                k, *_, v = param.split('=')
                self._params[k] = v
        return self._params

    @property
    def method(self) -> Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
        return self.scope['method']

    @property
    def path(self) -> str:
        return self.scope['path']

    @property
    def server(self) -> Address:
        return Address(*self.scope['server'])

    @property
    def client(self) -> Address:
        return Address(*self.scope['client'])

    @property
    def http_version(self) -> str:
        return self.scope['http_version']

    @property
    def scheme(self) -> str:
        return self.scope['scheme']

    @property
    def data(self):
        """Return The Validated Data
        It has been set on API.validate_input() while request is happening
        """
        return self._validated_data

    @property
    def pure_data(self) -> dict:
        """This is the data before validation"""
        from panther.logger import logger

        if self._data is None:

            body = self._body.decode('utf-8', errors='replace') or {}
            if self.headers.content_type is None:
                self._data = body
            elif self.headers.content_type == 'application/json':
                self._data = json.loads(body)
            elif self.headers.content_type[:19] == 'multipart/form-data':
                self._data = read_multipart_form_data(content_type=self.headers.content_type, body=body)
            else:
                logger.error(f'{self.headers.content_type} Is Not Supported.')
                self._data = {}

        return self._data

    def set_validated_data(self, validated_data) -> None:
        self._validated_data = validated_data

    @property
    def user(self):
        return self._user

    def set_user(self, user) -> None:
        self._user = user



from datetime import datetime

from panther import version
from panther.utils import generate_secret_key

apis_py = """from datetime import datetime

from panther.app import API
from panther.configs import config
from panther import version, status
from panther.request import Request
from panther.response import Response


@API()
async def hello_world():
    return {'detail': 'Hello World'}


@API(cache=True)
async def info(request: Request):
    data = {
        'version': version(),
        'datetime_now': datetime.now().isoformat(),
        'user_agent': request.headers.user_agent,
        'db_engine': config['db_engine'],
    }
    return Response(data=data, status_code=status.HTTP_202_ACCEPTED)
"""

models_py = """from panther.db import Model
"""

serializers_py = """from pydantic import BaseModel as Serializer
"""

app_urls_py = """from app.apis import hello_world, info

urls = {
    '': hello_world,
    'info/': info,
}
"""

configs_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""

from pathlib import Path
from panther.utils import load_env


BASE_DIR = Path(__name__).resolve().parent
env = load_env(BASE_DIR / '.env')

DB_NAME = env['DB_NAME']
SECRET_KEY = env['SECRET_KEY']


MIDDLEWARES = [
    ('panther.middlewares.db.Middleware', {'url': f'pantherdb://{BASE_DIR}/{DB_NAME}.pantherdb'}),
]

USER_MODEL = 'panther.db.models.User'

MONITORING = True

LOG_QUERIES = True

URLs = 'core/urls.py'
""" % datetime.now().date().isoformat()

middlewares = """from panther.middlewares import BaseMiddleware
"""

env = """
SECRET_KEY = '%s'

DB_NAME = '{PROJECT_NAME}'
""" % generate_secret_key()

main_py = """from panther import Panther

app = Panther(__name__)
"""

urls_py = """from app.urls import urls as app_urls

urls = {
    '/': app_urls, 
}
"""

git_ignore = """__pycache__/
.venv/
.idea/
.env
logs/
*.pantherdb
"""

requirements = """panther==%s
""" % version()

Template = {
    'app': {
        'apis.py': apis_py,
        'models.py': models_py,
        'serializers.py': serializers_py,
        'urls.py': app_urls_py,
    },
    'core': {
        'configs.py': configs_py,
        'middlewares.py': middlewares,
        'urls.py': urls_py,
    },
    'main.py': main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}



import os
import re

from setuptools import setup


def panther_version() -> str:
    with open(os.path.join('panther/__init__.py')) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


VERSION = panther_version()
DESCRIPTION = open('README.md').read()

EXTRAS_REQUIRE = {
    'full': [
        'python-jose>=3.3.0',
        'pymongo>=4.3.3',
    ]
}

setup(
    name='panther',
    version=VERSION,
    python_requires='>=3.11',
    author='Ali RajabNezhad',
    author_email='alirn76@yahoo.com',
    url='https://github.com/alirn76/panther',
    description='Fast &  Friendly, Web Framework For Building Async APIs',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
    ],
    entry_points={
        'console_scripts': ['panther=panther.cli.main:start'],
    },
    package_data={
        'panther': ['cli/*'],
    },
    install_requires=[
        'pantherdb>=1.2.0',
        'pydantic>=1.10.5',
        'watchfiles>=0.18.1',
        'uvicorn[standard]',
        'bpython>=0.24',
        'bson>=0.5.10',
        'redis>=4.0.2',
        'rich>=13.3.1',
    ],
    extras_require=EXTRAS_REQUIRE,
)



import bson
from pydantic import Field
from pydantic.main import BaseModel as PydanticBaseModel

from panther.configs import config
from panther.db.queries import Query


class BsonObjectId(bson.ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, str):
            try:
                bson.ObjectId(v)
            except Exception:
                raise TypeError('Invalid ObjectId')
        elif not isinstance(v, bson.ObjectId):
            raise TypeError('ObjectId required')
        return str(v)


if config['db_engine'] == 'pantherdb':
    IDType = int
else:
    IDType = BsonObjectId


class Model(PydanticBaseModel, Query):
    id: IDType | None = Field(alias='_id')

    @property
    def _id(self):
        if IDType is int:
            return self.id
        else:
            return bson.ObjectId(self.id) if self.id else None


class User(Model):
    first_name: str | None
    last_name: str | None



from panther.db.connection import DBSession
from panther.middlewares.base import BaseMiddleware
from panther.request import Request
from panther.response import Response


class Middleware(BaseMiddleware):

    def __init__(self, **kwargs):
        self.url = kwargs['url']

    async def before(self, request: Request) -> Request:
        self.db = DBSession(db_url=self.url)
        return request

    async def after(self, response: Response) -> Response:
        self.db.close()
        return response
